use serde::{Deserialize, Serialize};
use strum::Display;

use subxt::utils::H256;
use subxt::config::substrate::DigestItem;

use crate::types::{
    ActionLevel, CasperExtrinsicDetails, EraInfo, Nominator, PeerInformation, SessionKeyInfo, SystemAccount
};

#[derive(Debug, Clone, PartialEq, Eq, Display, Serialize, Deserialize)]
pub enum Action {
    Tick,
    Render,
    Resize(u16, u16),
    Suspend,
    Resume,
    Quit,
    ClearScreen,
    Error(String),
    Help,
    CheckPendingTransactions,

    SetMode(crate::app::Mode),
    SetActiveScreen(crate::app::Mode),
    UsedExplorerBlock(Option<u32>),
    UsedExplorerLog(Option<String>),
    UsedAccount(String),
    NewAccount(String),
    NewAddressBookRecord(String, String),

    ClosePopup,

    BalanceRequest([u8; 32], bool),
    BalanceResponse([u8; 32], Option<SystemAccount>),
    BalanceSetActive(Option<SystemAccount>),

    RenameAccount(String),
    RenameAddressBookRecord(String),
    UpdateAccountName(String),
    UpdateAddressBookRecord(String),
    TransferTo(String),

    TransferBalance(String, [u8; 32], u128),
    WalletLog(String, ActionLevel),
    ValidatorLog(String, ActionLevel),

    NewBestBlock(u32),
    NewBestHash(H256),
    NewFinalizedBlock(u32),
    NewFinalizedHash(H256),
    BestBlockUpdated(u32),
    ExtrinsicsLength(u32, usize),
    ValidatorsNumber(u32),
    NominatorsNumber(u32),
    Inflation(String),
    Apy(String),

    GetBlockAuthor(H256, Vec<DigestItem>),
    SetBlockAuthor(H256, String),

    GetNodeName,
    GetSystemHealth,
    GetGenesisHash,
    GetChainName,
    GetChainVersion,
    GetPendingExtrinsics,
    GetConnectedPeers,
    GetSessionKeys([u8; 32]),
    GetQueuedSessionKeys([u8; 32]),
    GetListenAddresses,
    GetLocalIdentity,

    GetLatestBlock,
    GetFinalizedBlock,
    GetActiveEra,
    GetCurrentEra,
    GetEpochProgress,
    GetValidatorsNumber,
    GetNominatorsNumber,
    GetInflation,
    GetNominatorsByValidator([u8; 32]),
    GetValidatorAllRewards([u8; 32]),
    GetValidatorLedger([u8; 32]),
    GetIsStashBonded([u8; 32]),
    GetErasStakersOverview([u8; 32]),
    GetValidatorPrefs([u8; 32]),

    SetNodeName(Option<String>),
    SetSystemHealth(Option<usize>, bool, bool),
    SetGenesisHash(Option<H256>),
    SetChainName(Option<String>),
    SetChainVersion(Option<String>),
    SetStashAccount([u8; 32]),

    BestBlockInformation(H256, u32),
    FinalizedBlockInformation(H256, u32),
    ExtrinsicsForBlock(u32, Vec<CasperExtrinsicDetails>),
    SetActiveEra(EraInfo),
    SetCurrentEra(u32),
    SetEpochProgress(u64, u64),
    SetPendingExtrinsicsLength(usize), 
    SetConnectedPeers(Vec<PeerInformation>),
    SetSessionKey(String, SessionKeyInfo),
    SetListenAddresses(Vec<String>),
    SetLocalIdentity(String),
    SetNominatorsByValidator(Vec<Nominator>),
    SetValidatorEraReward(u32, u128),
    SetValidatorEraClaimed(u32, bool),
    SetValidatorEraSlash(u32, u128),
    SetValidatorEraUnlocking(u32, u128),
    SetBondedAmount(bool),
    SetStakedAmountRatio(u128, u128),
    SetStakedRatio(u128, u128),
    SetValidatorPrefs(u32, bool),

    GetTotalIssuance,
    GetExistentialDeposit,

    SetExistentialDeposit(u128),
    SetTotalIssuance(Option<u128>),
}
