use std::time::Instant;

use color_eyre::Result;
use ratatui::{
    layout::{Alignment, Rect}, 
    widgets::{Block, Padding, Paragraph, Wrap}, 
    Frame
};

use super::{Component, PartialComponent, CurrentTab};
use crate::{
    config::Config,
    widgets::{BigText, PixelSize}, 
    action::Action, 
    palette::StylePalette,
};

#[derive(Debug)]
pub struct BlockTicker {
    last_block: u32,
    last_block_time: Instant,
    palette: StylePalette
}

impl Default for BlockTicker {
    fn default() -> Self {
        Self::new()
    }
}

impl BlockTicker {
    pub fn new() -> Self {
        Self {
            last_block: 0,
            last_block_time: Instant::now(),
            palette: StylePalette::default(),
        }
    }

    fn block_found(&mut self, block: u32) -> Result<()> {
        if self.last_block < block {
            self.last_block_time = Instant::now();
            self.last_block = block;
        }
        Ok(())
    }
}

impl PartialComponent for BlockTicker {
    fn set_active(&mut self, _current_tab: CurrentTab) {}
}

impl Component for BlockTicker {
    fn register_config_handler(&mut self, config: Config) -> Result<()> {
        if let Some(style) = config.styles.get(&crate::app::Mode::Explorer) {
            self.palette.with_normal_style(style.get("normal_style").copied());
            self.palette.with_hover_style(style.get("hover_style").copied());
            self.palette.with_normal_border_style(style.get("normal_border_style").copied());
            self.palette.with_hover_border_style(style.get("hover_border_style").copied());
            self.palette.with_normal_title_style(style.get("normal_title_style").copied());
            self.palette.with_hover_title_style(style.get("hover_title_style").copied());
        }
        Ok(())
    }

    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        match action {
            Action::BestBlockUpdated(block) => self.block_found(block)?,
            _ => {}
        };
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let [place, _, _] = super::explorer_block_info_layout(area);

        let passed = (Instant::now() - self.last_block_time).as_secs_f64();
        let text = if passed < 60.0 {
            format!("{:.1}s", passed)
        } else if passed < 3600.0 {
            let passed = passed % 3600.0 / 60.0;
            format!("{:.1}m", passed)
        } else {
            let passed = passed % (24.0 * 3600.0) / 3600.0;
            format!("{:.1}h", passed)
        };

        let (border_style, border_type) = self.palette.create_border_style(false);
        let height = place.as_size().height;
        let width = place.as_size().width;
        let text_width = text.len() as u16 * 4 + 2;

        if width < text_width || height < 7 {
            let paragraph = Paragraph::new(text)
                .block(Block::bordered()
                    .border_style(border_style)
                    .border_type(border_type)
                    .title_alignment(Alignment::Right)
                    .title_style(self.palette.create_title_style(false))
                    .padding(Padding::new(0, 0, height.saturating_sub(2) / 2, 0))
                    .title("Passed"))
                .alignment(Alignment::Center)
                .wrap(Wrap { trim: true });
            frame.render_widget(paragraph, place);
        } else {
            let big_text = BigText::builder()
                .centered()
                .pixel_size(PixelSize::Quadrant)
                .style(self.palette.create_basic_style(false))
                .lines(vec![
                    text.into(),
                ])
                .build();
            let paragraph = Paragraph::new("")
                .block(Block::bordered()
                    .border_style(border_style)
                    .border_type(border_type)
                    .title_alignment(Alignment::Right)
                    .title_style(self.palette.create_title_style(false))
                    .title("Passed"))
                .alignment(Alignment::Center)
                .wrap(Wrap { trim: true });
            frame.render_widget(paragraph, place);

            let height_offset = height.saturating_sub(2) / 2;
            let place = place.offset(ratatui::layout::Offset { x: 1, y: height_offset as i32 })
                .intersection(place);
            frame.render_widget(big_text, place);
        }
        Ok(())
    }
}
