use color_eyre::Result;
use ratatui::layout::Constraint;
use ratatui::{
    text::Text,
    layout::{Alignment, Rect}, 
    widgets::{Block, Cell, Row, Table}, 
    Frame
};

use super::{PartialComponent, Component, CurrentTab};
use crate::{
    action::Action, 
    config::Config, 
    palette::StylePalette, 
};

pub struct RewardDetails {
    palette: StylePalette,
    commission: u32,
    nominators_blocked: bool,
    apy: String,
    inflation: String,
}

impl Default for RewardDetails {
    fn default() -> Self {
        Self::new()
    }
}

impl RewardDetails {
    pub fn new() -> Self {
        Self {
            palette: StylePalette::default(),
            commission: 0,
            nominators_blocked: false,
            apy: String::from("0.0%"),
            inflation: String::from("0.0%"),
        }
    }

    fn comission_to_string(&self) -> String {
        if self.nominators_blocked {
            "blocked".to_string()
        } else {
            let result = self.commission as f64 / 1_000_000_000.0;
            format!("{:.1}%", result)
        }
    }
}

impl PartialComponent for RewardDetails {
    fn set_active(&mut self, _current_tab: CurrentTab) { }
}

impl Component for RewardDetails {
    fn register_config_handler(&mut self, config: Config) -> Result<()> {
        if let Some(style) = config.styles.get(&crate::app::Mode::Validator) {
            self.palette.with_normal_style(style.get("normal_style").copied());
            self.palette.with_hover_style(style.get("hover_style").copied());
            self.palette.with_normal_border_style(style.get("normal_border_style").copied());
            self.palette.with_hover_border_style(style.get("hover_border_style").copied());
            self.palette.with_normal_title_style(style.get("normal_title_style").copied());
            self.palette.with_hover_title_style(style.get("hover_title_style").copied());
            self.palette.with_highlight_style(style.get("highlight_style").copied());
            self.palette.with_scrollbar_style(style.get("scrollbar_style").copied());
        }
        Ok(())
    }

    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        match action {
            Action::SetValidatorPrefs(commission, disabled) => {
                self.commission = commission;
                self.nominators_blocked = disabled;
            }
            Action::Apy(apy) => self.apy = apy,
            Action::Inflation(inflation) => self.inflation = inflation,
            _ => {}
        };
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let [_, _, place] = super::validator_balance_layout(area);
        let (border_style, border_type) = self.palette.create_border_style(false);

        let table = Table::new(
            vec![
                Row::new(vec![
                    Cell::from(Text::from("Nominators".to_string()).alignment(Alignment::Left)),
                    Cell::from(Text::from(self.comission_to_string()).alignment(Alignment::Right)), 
                ]),
                Row::new(vec![
                    Cell::from(Text::from("Current APY".to_string()).alignment(Alignment::Left)),
                    Cell::from(Text::from(self.apy.clone()).alignment(Alignment::Right)), 
                ]),
                Row::new(vec![
                    Cell::from(Text::from("Inflation".to_string()).alignment(Alignment::Left)),
                    Cell::from(Text::from(self.inflation.clone()).alignment(Alignment::Right)), 
                ]),
            ],
            [
                Constraint::Min(11),
                Constraint::Min(0),
            ],
        )
        .highlight_style(self.palette.create_highlight_style())
        .column_spacing(1)
        .block(Block::bordered()
            .border_style(border_style)
            .border_type(border_type)
            .title_alignment(Alignment::Right)
            .title_style(self.palette.create_title_style(false))
            .title("Reward details"));

        frame.render_widget(table, place);

        Ok(())
    }
}
