use subxt::ext::sp_runtime::Perbill;

// generated outside, based on params
// MIN_INFLATION:   u32 = 0_025_000;
// MAX_INFLATION:   u32 = 0_100_000;
// IDEAL_STAKE:     u32 = 0_750_000;
// FALLOFF:         u32 = 0_050_000;
// MAX_PIECE_COUNT: u32 = 40;
const PIECEWISE_LINEAR_POUNTS: [(Perbill, Perbill); 33] = [
    (Perbill::from_parts(0),          Perbill::from_parts(25000000)),
    (Perbill::from_parts(750000000),  Perbill::from_parts(100000000)),
    (Perbill::from_parts(758333000),  Perbill::from_parts(91817000)),
    (Perbill::from_parts(766666000),  Perbill::from_parts(84528000)),
    (Perbill::from_parts(774999000),  Perbill::from_parts(78033000)),
    (Perbill::from_parts(783331000),  Perbill::from_parts(72248000)),
    (Perbill::from_parts(791663000),  Perbill::from_parts(67094000)),
    (Perbill::from_parts(799996000),  Perbill::from_parts(62502000)),
    (Perbill::from_parts(808328000),  Perbill::from_parts(58411000)),
    (Perbill::from_parts(816661000),  Perbill::from_parts(54766000)),
    (Perbill::from_parts(824993000),  Perbill::from_parts(51519000)),
    (Perbill::from_parts(833325000),  Perbill::from_parts(48626000)),
    (Perbill::from_parts(841656000),  Perbill::from_parts(46049000)),
    (Perbill::from_parts(849988000),  Perbill::from_parts(43753000)),
    (Perbill::from_parts(858321000),  Perbill::from_parts(41707000)),
    (Perbill::from_parts(866651000),  Perbill::from_parts(39885000)),
    (Perbill::from_parts(874984000),  Perbill::from_parts(38261000)),
    (Perbill::from_parts(883313000),  Perbill::from_parts(36815000)),
    (Perbill::from_parts(891646000),  Perbill::from_parts(35526000)),
    (Perbill::from_parts(899976000),  Perbill::from_parts(34378000)),
    (Perbill::from_parts(908308000),  Perbill::from_parts(33355000)),
    (Perbill::from_parts(916636000),  Perbill::from_parts(32444000)),
    (Perbill::from_parts(924968000),  Perbill::from_parts(31632000)),
    (Perbill::from_parts(933295000),  Perbill::from_parts(30909000)),
    (Perbill::from_parts(941619000),  Perbill::from_parts(30265000)),
    (Perbill::from_parts(949946000),  Perbill::from_parts(29691000)),
    (Perbill::from_parts(958265000),  Perbill::from_parts(29180000)),
    (Perbill::from_parts(966598000),  Perbill::from_parts(28724000)),
    (Perbill::from_parts(974925000),  Perbill::from_parts(28318000)),
    (Perbill::from_parts(983258000),  Perbill::from_parts(27956000)),
    (Perbill::from_parts(991578000),  Perbill::from_parts(27634000)),
    (Perbill::from_parts(999899000),  Perbill::from_parts(27347000)),
    (Perbill::from_parts(1000000000), Perbill::from_parts(27343000)),
];
const MAXIMUM_INFLATION: Perbill = Perbill::from_parts(100000000);

pub fn calculate_for_fraction(n: u128, d: u128) -> (Perbill, Perbill) {
    let n = n.min(d.clone());

    if PIECEWISE_LINEAR_POUNTS.is_empty() {
        return (MAXIMUM_INFLATION, Perbill::zero())
    }

    let next_point_index = PIECEWISE_LINEAR_POUNTS.iter().position(|p| n < p.0 * d.clone());

    let (prev, next) = if let Some(next_point_index) = next_point_index {
        if let Some(previous_point_index) = next_point_index.checked_sub(1) {
            (PIECEWISE_LINEAR_POUNTS[previous_point_index], PIECEWISE_LINEAR_POUNTS[next_point_index])
        } else {
            // There is no previous points, take first point ordinate
            let fraction = PIECEWISE_LINEAR_POUNTS.first().map(|p| p.1).unwrap_or_else(Perbill::zero);
            return (MAXIMUM_INFLATION, fraction)
        }
    } else {
        // There is no next points, take last point ordinate
        let fraction = PIECEWISE_LINEAR_POUNTS.last().map(|p| p.1).unwrap_or_else(Perbill::zero);
        return (MAXIMUM_INFLATION, fraction)
    };

    let delta_y = multiply_by_rational_saturating(
        abs_sub(n.clone(), prev.0 * d.clone()),
        abs_sub(next.1.deconstruct(), prev.1.deconstruct()),
        // Must not saturate as prev abscissa > next abscissa
        next.0.deconstruct().saturating_sub(prev.0.deconstruct()),
    );

    // If both subtractions are same sign then result is positive
    let fraction = if (n > prev.0 * d.clone()) == (next.1.deconstruct() > prev.1.deconstruct()) {
        (prev.1 * d).saturating_add(delta_y)
    } else {
        // Otherwise result is negative
        (prev.1 * d).saturating_sub(delta_y)
    };

    (MAXIMUM_INFLATION, Perbill::from_rational(fraction, d))
}

fn abs_sub<N: Ord + core::ops::Sub<Output = N> + Clone>(a: N, b: N) -> N where {
	a.clone().max(b.clone()) - a.min(b)
}

fn multiply_by_rational_saturating(value: u128, p: u32, q: u32) -> u128 {
	let q = q.max(1);
	let result_divisor_part = (value / q as u128).saturating_mul(p as u128);
	let result_remainder_part = {
		let rem = value % q as u128;
		let rem_u32 = rem as u32;
		let rem_part = rem_u32 as u64 * p as u64 / q as u64;
		rem_part as u128
	};
	result_divisor_part.saturating_add(result_remainder_part)
}

pub fn prepare_perbill_fraction_string(value: Perbill) -> String {
    let d = value.deconstruct();
    let mut m = 10_000_000;

    let units = d / m;
    let rest = d % m;

    for _ in 0..2 {
        m /= 10;
    }

    format!("{}.{}%", units, rest / m)
}
